#!/usr/bin/env python3

"""
APRS message data types.
"""

from enum import Enum

class APRSDataType(Enum):
    """
    APRS message types, given as the first byte in the information field,
    not including unused or reserved types.  Page 17 of APRS 1.0.1 spec.
    """
    MIC_E_BETA0         = 0x1c
    MIC_E_OLD_BETA0     = 0x1d
    POSITION            = ord('!')
    PEET_BROS_WX1       = ord('#')
    RAW_GPRS_ULT2K      = ord('$')
    AGRELO_DFJR         = ord('%')
    RESERVED_MAP        = ord('&')
    MIC_E_OLD           = ord("'")
    ITEM                = ord(')')
    PEET_BROS_WX2       = ord('*')
    TEST_DATA           = ord(',')
    POSITION_TS         = ord('/')
    MESSAGE             = ord(':')
    OBJECT              = ord(';')
    STATIONCAP          = ord('<')
    POSITION_MSGCAP     = ord('=')
    STATUS              = ord('>')
    QUERY               = ord('?')
    POSITION_TS_MSGCAP  = ord('@')
    TELEMETRY           = ord('T')
    MAIDENHEAD          = ord('[')
    WX                  = ord('_')
    MIC_E               = ord('`')
    USER_DEFINED        = ord('{')
    THIRD_PARTY         = ord('}')



#!/usr/bin/env python3

"""
KISS command unit tests
"""

from aioax25.kiss import KISSCommand, KISSCmdReturn, KISSCmdData
from nose.tools import eq_

def test_stuff_bytes():
    """
    Test that bytes are correctly escaped.
    """
    eq_(
            bytes(KISSCommand._stuff_bytes(
                b'A test frame.\n'
                b'FEND becomes FESC TFEND: \xc0\n'
                b'while FESC becomes FESC TFESC: \xdb\n'
            )),
            # This should be decoded as the following:
            b'A test frame.\n'
            b'FEND becomes FESC TFEND: \xdb\xdc\n'
            b'while FESC becomes FESC TFESC: \xdb\xdd\n'
    )

def test_unstuff_bytes():
    """
    Test that bytes are correctly unescaped.
    """
    eq_(
            bytes(KISSCommand._unstuff_bytes(
                b'A test frame.\n'
                b'If we see FESC TFEND, we should get FEND: \xdb\xdc\n'
                b'while if we see FESC TFESC, we should get FESC: \xdb\xdd\n'
                b'FESC followed by any other byte should yield those\n'
                b'two original bytes: \xdb\xaa \xdb\xdb \xdb\xdb\xdd\n'
            )),
            # This should unstuff to the following
            b'A test frame.\n'
            b'If we see FESC TFEND, we should get FEND: \xc0\n'
            b'while if we see FESC TFESC, we should get FESC: \xdb\n'
            b'FESC followed by any other byte should yield those\n'
            b'two original bytes: \xdb\xaa \xdb\xdb \xdb\xdb\n'
    )

def test_decode_unknown():
    """
    Test unknown KISS frames are decoded to the base KISSCommand base class.
    """
    frame = KISSCommand.decode(
            b'\x58unknown command payload'
    )
    assert isinstance(frame, KISSCommand)
    eq_(frame.cmd, 8)
    eq_(frame.port, 5)
    eq_(frame.payload, b'unknown command payload')

def test_decode_data():
    """
    Test the DATA frame is decoded correctly.
    """
    frame = KISSCommand.decode(b'\x90this is a data frame')
    assert isinstance(frame, KISSCmdData)
    eq_(frame.payload, b'this is a data frame')

def test_encode_unknown():
    """
    Test we can encode an arbitrary frame.
    """
    eq_(bytes(KISSCommand(port=3, cmd=12, payload=b'testing')),
            b'\x3ctesting')

def test_encode_return():
    """
    Test we can encode a RETURN frame
    """
    eq_(bytes(KISSCmdReturn()), b'\xff')

def test_encode_data():
    """
    Test we can encode a DATA frame
    """
    eq_(bytes(KISSCmdData(port=2, payload=b'a frame')), b'\x20a frame')



#!/usr/bin/env python3

"""
Serial KISS interface unit tests.
"""

from aioax25 import kiss
import time
from serial import EIGHTBITS, PARITY_NONE, STOPBITS_ONE

from nose.tools import eq_, assert_less, assert_in
from ..loop import DummyLoop

class DummySerial(object):
    FILENO = 123

    def __init__(self, port, baudrate, bytesize, parity, stopbits,
            timeout, xonxoff, rtscts, write_timeout, dsrdtr,
            inter_byte_timeout):

        eq_(port, '/dev/ttyS0')
        eq_(baudrate, 9600)
        eq_(bytesize, EIGHTBITS)
        eq_(parity, PARITY_NONE)
        eq_(stopbits, STOPBITS_ONE)
        eq_(timeout, None)
        eq_(xonxoff, False)
        eq_(rtscts, False)
        eq_(write_timeout, None)
        eq_(dsrdtr, False)
        eq_(inter_byte_timeout, None)

        self.rx_buffer = bytearray()
        self.tx_buffer = bytearray()
        self.flushes = 0
        self.closed = False
        self.read_exception = None

    def fileno(self):
        return self.FILENO

    def flush(self):
        self.flushes += 1

    def write(self, data):
        self.tx_buffer += data

    def read(self, length):
        if self.read_exception is not None:
            raise self.read_exception

        data = self.rx_buffer[0:length]
        self.rx_buffer = self.rx_buffer[length:]
        return data

    def close(self):
        self.closed = True

    @property
    def in_waiting(self):
        return len(self.rx_buffer)


# Stub the serial port class
kiss.Serial = DummySerial


def test_open():
    """
    Test we can open the port.
    """
    loop = DummyLoop()
    kissdev = kiss.SerialKISSDevice(
            device='/dev/ttyS0', baudrate=9600,
            loop=loop
    )
    assert kissdev._serial is None

    kissdev.open()

    # kissdev._serial should now be a DummySerial instance
    assert isinstance(kissdev._serial, DummySerial)

    # A reader should have been added
    assert_in(kissdev._serial.fileno(), loop.readers)
    eq_(loop.readers[DummySerial.FILENO], kissdev._on_recv_ready)

    # A call to initialise the device should have been scheduled.
    eq_(len(loop.calls), 1)
    (calltime, callfunc) = loop.calls.pop(0)
    assert_less(time.monotonic() - calltime, 0.01)
    eq_(callfunc, kissdev._init_kiss)


def test_close():
    """
    Test we can close the port.
    """
    loop = DummyLoop()
    kissdev = kiss.SerialKISSDevice(
            device='/dev/ttyS0', baudrate=9600,
            loop=loop, reset_on_close=False
    )

    # Force the port open
    kissdev._state = kiss.KISSDeviceState.OPEN
    serial = DummySerial(port='/dev/ttyS0', baudrate=9600,
            bytesize=EIGHTBITS, parity=PARITY_NONE, stopbits=STOPBITS_ONE,
            timeout=None, xonxoff=False, rtscts=False, write_timeout=None,
            dsrdtr=False, inter_byte_timeout=None)
    kissdev._serial = serial
    
    # Add a reader
    loop.readers[DummySerial.FILENO] = lambda *a, **k : None

    # Now try closing the port
    kissdev.close()

    # The reader should be gone
    eq_(len(loop.readers), 0)

    # The port should have been flushed
    eq_(serial.flushes, 1)

    # The port should be closed
    eq_(serial.closed, True)

    # The device should not reference the port
    eq_(kissdev._serial, None)

    # The port should now be in the closed state
    eq_(kissdev._state, kiss.KISSDeviceState.CLOSED)

def test_on_recv_ready():
    """
    Test _on_recv_ready reads from serial device and passes to _receive.
    """
    loop = DummyLoop()
    kissdev = kiss.SerialKISSDevice(
            device='/dev/ttyS0', baudrate=9600,
            loop=loop
    )
    assert kissdev._serial is None

    kissdev.open()
    kissdev._serial.rx_buffer += b'a test frame'

    # Kick the handler
    kissdev._on_recv_ready()

    # We should now see these bytes in the buffer
    eq_(bytes(kissdev._rx_buffer), b'a test frame')

def test_on_recv_ready_exception():
    """
    Test _on_recv_ready handles exceptions.
    """
    loop = DummyLoop()
    kissdev = kiss.SerialKISSDevice(
            device='/dev/ttyS0', baudrate=9600,
            loop=loop
    )
    assert kissdev._serial is None

    kissdev.open()
    kissdev._serial.read_exception = IOError('Whoopsie')
    kissdev._serial.rx_buffer += b'a test frame'

    # Kick the handler -- there should be no error raised
    kissdev._on_recv_ready()

def test_send_raw_data():
    """
    Test _send_raw_data passes the data to the serial device.
    """
    loop = DummyLoop()
    kissdev = kiss.SerialKISSDevice(
            device='/dev/ttyS0', baudrate=9600,
            loop=loop
    )
    assert kissdev._serial is None

    kissdev.open()
    kissdev._send_raw_data(b'a test frame')
    eq_(bytes(kissdev._serial.tx_buffer), b'a test frame')



#!/usr/bin/env python3

"""
AX.25 Interface handler
"""

import logging
import asyncio
import random
from functools import partial
from .signal import Signal
import re

from .frame import AX25Frame
from .router import Router


class AX25Interface(Router):
    """
    The AX25Interface class represents a logical AX.25 interface.
    The interface handles basic queueing and routing of message traffic.

    Outgoing messages are queued and sent when there is a break of greater
    than the cts_delay (250ms) + a randomisation factor (cts_rand).
    Messages may be cancelled prior to transmission.
    """

    def __init__(self, kissport, cts_delay=0.25,
            cts_rand=0.25, log=None, loop=None):
        # Initialise the superclass
        super(AX25Interface, self).__init__()

        if log is None:
            log = logging.getLogger(self.__class__.__module__)

        if loop is None:
            loop = asyncio.get_event_loop()

        self._log = log
        self._loop = loop
        self._port = kissport

        # Message queue
        self._tx_queue = []
        self._tx_pending = None

        # Clear-to-send delay and randomisation factor
        self._cts_delay = cts_delay
        self._cts_rand = cts_rand

        # Clear-to-send expiry
        self._cts_expiry = loop.time() \
                + cts_delay \
                + (random.random() * cts_rand)

        # Bind to the KISS port to receive raw messages.
        kissport.received.connect(self._on_receive)

    def transmit(self, frame, callback=None):
        """
        Enqueue a message for transmission.  Optionally give a call-back
        function to receive notification of transmission.
        """
        self._log.debug('Adding to queue: %s', frame)
        self._tx_queue.append((frame, callback))
        if not self._tx_pending:
            self._schedule_tx()

    def cancel_transmit(self, frame):
        """
        Cancel the transmission of a frame.
        """
        self._log.debug('Removing from queue: %s', frame)
        self._tx_queue = list(filter(
            lambda item : item[0] is not frame,
            self._tx_queue
        ))

    def _reset_cts(self):
        """
        Reset the clear-to-send timer.
        """
        cts_expiry = self._loop.time() \
                + self._cts_delay + (random.random() * self._cts_rand)

        # Ensure CTS expiry never goes backwards!
        while cts_expiry < self._cts_expiry:
            cts_expiry += (random.random() * self._cts_rand)
        self._cts_expiry = cts_expiry

        self._log.debug('Clear-to-send expiry at %s', self._cts_expiry)
        if self._tx_pending:
            # We were waiting for a clear-to-send, so re-schedule.
            self._schedule_tx()

    def _on_receive(self, frame):
        """
        Handle an incoming message.
        """
        self._reset_cts()
        super(AX25Interface, self)._on_receive(frame)

    def _schedule_tx(self):
        """
        Schedule the transmit timer to take place after the CTS expiry.
        """
        if self._tx_pending:
            self._tx_pending.cancel()

        delay = self._cts_expiry - self._loop.time()
        if delay > 0:
            self._log.debug('Scheduling next transmission in %s', delay)
            self._tx_pending = self._loop.call_later(delay, self._tx_next)
        else:
            self._log.debug('Scheduling next transmission ASAP')
            self._tx_pending = self._loop.call_soon(self._tx_next)

    def _tx_next(self):
        """
        Transmit the next message.
        """
        self._tx_pending = None

        try:
            (frame, callback) = self._tx_queue.pop(0)
        except IndexError:
            self._log.debug('No traffic to transmit')
            return

        try:
            self._log.debug('Transmitting %s', frame)
            self._port.send(frame)
            if callback:
                self._log.debug('Notifying sender of %s', frame)
                self._loop.call_soon(partial(
                    callback,
                    interface=self,
                    frame=frame
                ))
        except:
            self._log.exception('Failed to transmit %s', frame)

        self._reset_cts()
        if len(self._tx_queue) > 0:
            self._schedule_tx()



#!/usr/bin/env python3

"""
KISS serial interface handler.  This allows basic support for talking to
KISS-based TNCs, managing the byte stuffing/unstuffing.
"""

from enum import Enum
from asyncio import get_event_loop
from serial import Serial, EIGHTBITS, PARITY_NONE, STOPBITS_ONE
from .signal import Signal
from binascii import b2a_hex
import logging


# Constants


BYTE_FEND   = 0xc0
BYTE_FESC   = 0xdb
BYTE_TFEND  = 0xdc
BYTE_TFESC  = 0xdd

CMD_DATA    = 0x00
CMD_TXDELAY = 0x01
CMD_P       = 0x02
CMD_SLOTTIME= 0x03
CMD_TXTAIL  = 0x04
CMD_FDUPLEX = 0x05
CMD_SETHW   = 0x06
CMD_RETURN  = 0x0f


# States


class KISSDeviceState(Enum):
    """
    States permitted by a KISS device:
    - CLOSED: Serial port is closed
    - INIT: Serial port just opened, TNC may be in TNC2-mode and
            the library is attempting to put it into KISS mode.
    - OPEN: Serial port is open, TNC in KISS mode.
    - CLOSING: Close instruction just received.  Putting TNC back into
      TNC2-mode if requested then closing the port.
    """
    CLOSED = 0
    OPENING = 1
    OPEN = 2
    CLOSING = 3


# Command classes


class KISSCommand(object):
    """
    KISS base command class
    """

    # Known commands: this will be populated later.
    _KNOWN_COMMANDS = {}

    @classmethod
    def _register(cls, cmd, subclass):
        assert cmd not in cls._KNOWN_COMMANDS
        cls._KNOWN_COMMANDS[cmd] = subclass

    @classmethod
    def _stuff_bytes(cls, data):
        """
        Byte-stuff incoming byte string.
        """
        for byte in data:
            if byte == BYTE_FEND:
                yield BYTE_FESC
                yield BYTE_TFEND
            elif byte == BYTE_FESC:
                yield BYTE_FESC
                yield BYTE_TFESC
            else:
                yield byte

    @classmethod
    def _unstuff_bytes(cls, data):
        """
        Un-byte-stuff incoming byte string.
        """
        last = None
        for byte in data:
            if byte == BYTE_FESC:
                if last == BYTE_FESC:
                    yield last
                else:
                    last = BYTE_FESC
            elif last == BYTE_FESC:
                if byte == BYTE_TFEND:
                    yield BYTE_FEND
                elif byte == BYTE_TFESC:
                    yield BYTE_FESC
                else:
                    yield last
                    yield byte
                last = None
            else:
                yield byte

    @classmethod
    def decode(cls, frame):
        """
        Decode a raw KISS frame.
        """
        frame = bytearray(cls._unstuff_bytes(frame))
        port = frame[0] >> 4
        cmd  = frame[0] & 0x0f
        subclass = cls._KNOWN_COMMANDS.get(cmd, cls)
        return subclass(port=port, cmd=cmd, payload=frame[1:])

    def __init__(self, port, cmd, payload=None):
        self._port = port
        self._cmd = cmd
        self._payload = payload

    def __bytes__(self):
        out = bytearray([
            ((self._port & 0x0f) << 4) | (self._cmd & 0x0f)
        ])
        if self._payload:
            out += self._payload

        # Encode the byte sequences
        return bytes(self._stuff_bytes(out))

    def __str__(self):
        return '%s{Port %d, Cmd 0x%02x, Payload %s}' % (
                self.__class__.__name__,
                self.port,
                self.cmd,
                b2a_hex(self.payload).decode()
        )

    @property
    def port(self):
        return self._port

    @property
    def cmd(self):
        return self._cmd

    @property
    def payload(self):
        return self._payload


class KISSCmdReturn(KISSCommand):
    """
    Emit a return command to the TNC.
    """
    def __init__(self):
        super(KISSCmdReturn, self).__init__(port=15, cmd=CMD_RETURN)
KISSCommand._register(CMD_RETURN, KISSCmdReturn)


class KISSCmdData(KISSCommand):
    def __init__(self, port, payload, cmd=CMD_DATA):
        assert cmd == CMD_DATA
        super(KISSCmdData, self).__init__(port=port,
                cmd=CMD_DATA, payload=payload)
KISSCommand._register(CMD_DATA, KISSCmdData)


# KISS device interface


class BaseKISSDevice(object):
    """
    Base class for a KISS device.  This may have between 1 and 16 KISS
    ports hanging off it.
    """
    def __init__(self, reset_on_close=True,
            send_block_size=128, send_block_delay=0.1,
            log=None, loop=None):
        if log is None:
            log = logging.getLogger(self.__class__.__module__)
        if loop is None:
            loop = get_event_loop()
        self._log = log
        self._protocol = None
        self._rx_buffer = bytearray()
        self._tx_buffer = bytearray()
        self._loop = loop
        self._port = {}
        self._state = KISSDeviceState.CLOSED
        self._reset_on_close = reset_on_close
        self._send_block_size = send_block_size
        self._send_block_delay = send_block_delay

    def _receive(self, data):
        """
        Handle incoming data by appending to our receive buffer.  The
        data given may contain partial frames.
        """
        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug('RECV RAW %r', b2a_hex(data).decode())
        self._rx_buffer += data
        self._loop.call_soon(self._receive_frame)

    def _send(self, frame):
        """
        Send a frame via the underlying transport.
        """
        rawframe = bytes(frame)

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug('XMIT FRAME %r', b2a_hex(rawframe).decode())

        if not self._tx_buffer.endswith(bytearray([BYTE_FEND])):
            self._tx_buffer += bytearray([BYTE_FEND])

        self._tx_buffer += bytes(rawframe) \
                         + bytearray([BYTE_FEND])
        self._loop.call_soon(self._send_data)

    def _receive_frame(self):
        """
        Scan the receive buffer for incoming frames and send these to the
        underlying device.  If more than one frame is present, schedule
        ourselves again with the IO loop.
        """
        # Locate the first FEND byte
        try:
            start = self._rx_buffer.index(BYTE_FEND)
        except ValueError:
            # No frames waiting
            self._rx_buffer = bytearray()
            return

        self._log.debug('RECV FRAME start at %d', start)

        # Discard the proceeding junk
        self._rx_buffer = self._rx_buffer[start:]
        del start
        assert self._rx_buffer[0] == BYTE_FEND

        # Locate the last FEND byte of the frame
        try:
            end = self._rx_buffer.index(BYTE_FEND, 1)
        except ValueError:
            # Uhh huh, so frame is incomplete.
            return

        self._log.debug('RECV FRAME end at %d', end)

        # Everything between those points is our frame.
        frame = self._rx_buffer[1:end]
        self._rx_buffer = self._rx_buffer[end:]

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug('RECEIVED FRAME %s, REMAINING %s',
                    b2a_hex(frame).decode(),
                    b2a_hex(self._rx_buffer).decode()
            )

        # Two consecutive FEND bytes are valid, ignore these "empty" frames
        if len(frame) > 0:
            # Decode the frame
            self._loop.call_soon(self._dispatch_rx_frame,
                    KISSCommand.decode(frame))

        # If we just have a FEND, stop here.
        if bytes(self._rx_buffer) == bytearray([BYTE_FEND]):
            return

        # If there is more to send, call ourselves via the IO loop
        if len(self._rx_buffer):
            self._loop.call_soon(self._receive_frame)

    def _dispatch_rx_frame(self, frame):
        """
        Pass a frame to the underlying interface.
        """
        try:
            port = self._port[frame.port]
        except KeyError:
            # Port not defined.
            self._log.debug('RECV FRAME dropped %s', frame)
            return

        # Dispatch this frame to the port.  Swallow exceptions so we
        # don't choke the IO loop.
        self._log.debug('RECV FRAME dispatch %s', frame)
        try:
            port._receive_frame(frame)
        except:
            logging.getLogger(self.__class__.__module__).exception(
                    'Port %s failed to handle frame %s',
                    port, frame
            )

    def _send_data(self):
        """
        Send the next block of data waiting in the buffer.
        """
        data = self._tx_buffer[:self._send_block_size]
        self._tx_buffer = self._tx_buffer[self._send_block_size:]

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug('XMIT RAW %r', b2a_hex(data).decode())

        self._send_raw_data(data)

        # If we are closing, wait for this to be sent
        if (self._state == KISSDeviceState.CLOSING) and \
                (len(self._tx_buffer) == 0):
            self._close()
            return

        if self._tx_buffer:
            self._loop.call_later(self._send_block_delay, self._send_data)

    def _init_kiss(self):
        assert self.state == KISSDeviceState.OPENING, \
                'Device is not opening'
        # For now, just blindly send a INT KISS command followed by a RESET.
        self._send_raw_data(b'\rINT KISS\rRESET\r')
        self._state = KISSDeviceState.OPEN

    def __getitem__(self, port):
        """
        Retrieve an instance of a specified port.
        """
        try:
            return self._port[port]
        except KeyError:
            pass

        self._log.debug('OPEN new port %d', port)
        p = KISSPort(self, port, log=self._log.getChild('port%d' % port))
        self._port[port] = p
        return p

    @property
    def state(self):
        return self._state

    def open(self):
        assert self.state == KISSDeviceState.CLOSED, \
                'Device is not closed'
        self._log.debug('Opening device')
        self._state = KISSDeviceState.OPENING
        self._open()

    def close(self):
        assert self.state == KISSDeviceState.OPEN, \
                'Device is not open'
        self._log.debug('Closing device')
        self._state = KISSDeviceState.CLOSING
        if self._reset_on_close:
            self._send(KISSCmdReturn())
        else:
            self._close()


class SerialKISSDevice(BaseKISSDevice):
    def __init__(self, device, baudrate, *args, **kwargs):
        super(SerialKISSDevice, self).__init__(*args, **kwargs)
        self._serial = None
        self._device = device
        self._baudrate = baudrate

    def _open(self):
        self._serial = Serial(port=self._device, baudrate=self._baudrate,
                bytesize=EIGHTBITS, parity=PARITY_NONE, stopbits=STOPBITS_ONE,
                timeout=None, xonxoff=False, rtscts=False, write_timeout=None,
                dsrdtr=False, inter_byte_timeout=None)
        self._loop.add_reader(self._serial.fileno(), self._on_recv_ready)
        self._loop.call_soon(self._init_kiss)

    def _close(self):
        # Remove handlers
        self._loop.remove_reader(self._serial.fileno())

        # Wait for all data to be sent.
        self._serial.flush()

        # Close the port
        self._serial.close()
        self._serial = None
        self._state = KISSDeviceState.CLOSED

    def _on_recv_ready(self):
        try:
            self._receive(self._serial.read(self._serial.in_waiting))
        except:
            self._log.exception('Failed to read from serial device')

    def _send_raw_data(self, data):
        self._serial.write(data)


# Port interface


class KISSPort(object):
    """
    A KISS port represents a port interface on a KISS device.  There can be
    a maximum of 16 ports per device, identified by a 4-bit integer.
    """
    def __init__(self, device, port, log):
        """
        Create a new port attached to the given device.
        """
        self._device = device
        self._port = port
        self._log = log

        # Signal for receiving packets
        self.received = Signal()

    @property
    def port(self):
        return self._port

    def send(self, frame):
        """
        Send a raw AX.25 frame to the TNC via this port.
        """
        self._log.debug('XMIT AX.25 %s', frame)
        self._device._send(KISSCmdData(self.port, bytes(frame)))

    def _receive_frame(self, frame):
        """
        Receive and emit an incoming frame from the port.
        """
        self._log.debug('Received frame %s', frame)

        if not isinstance(frame, KISSCmdData):
            # TNC is not supposed to send this!
            return

        self.received.emit(frame=frame.payload)



#!/usr/bin/env python3

"""
APRS messaging router.
"""

from .message import APRSMessageFrame
from ..router import Router


class APRSRouter(Router):
    """
    Route a APRS message frame according to the addressee field, if any.
    """
    def _get_destination(self, frame):
        if isinstance(frame, APRSMessageFrame):
            return frame.addressee
        else:
            return super(APRSRouter, self)._get_destination(frame)



#!/usr/bin/env python3

"""
APRS Digipeating module
"""

import logging
import weakref
import re
from ..frame import AX25FrameHeader, AX25Address

# APRS WIDEn/TRACEn regular expression pattern
DIGI_RE = re.compile(r'^(WIDE|TRACE)(\d)$')

class APRSDigipeater(object):
    """
    The APRSDigipeater class implemenets a pure WIDEn-N style digipeater
    handler, hooking into the Router handling hooks and editing the
    digipeater path of all unique APRS messages seen.
    """

    def __init__(self, log=None):
        """
        Create a new digipeater module instance.
        """
        if log is None:
            log = logging.getLogger(self.__class__.__module__)
        self._log = log
        self._mydigi = set()

    @property
    def mydigi(self):
        """
        Return the set of digipeater calls and aliases this digi responds to.
        """
        return self._mydigi.copy()

    @mydigi.setter
    def mydigi(self, aliases):
        """
        Replace the list of digipeater calls and aliases this digi responds to.
        """
        self._mydigi = set([
            AX25Address.decode(call).normalised
            for call in aliases
        ])

    def addaliases(self, *aliases):
        """
        Add one or more aliases to the digipeater handler.
        """
        for call in aliases:
            self._mydigi.add(AX25Address.decode(call).normalised)

    def rmaliases(self, *aliases):
        """
        Remove one or more aliases from the digipeater handler.
        """
        for call in aliases:
            self._mydigi.discard(AX25Address.decode(call).normalised)

    def connect(self, aprsint, addcall=True):
        """
        Connect to an APRS interface.  This hooks the received_msg signal
        to receive (de-duplicated) incoming traffic and adds the APRS
        interface's call-sign/SSID to the "mydigi" list.

        Note that a message is digipeated on the interface it was received
        *ONLY*.  Cross-interface digipeating is not implemented at this time.
        """
        self._log.debug('Connecting to %s (add call %s)', aprsint, addcall)
        aprsint.received_msg.connect(self._on_receive)
        if addcall:
            self.addaliases(aprsint.mycall)

    def disconnect(self, aprsint, rmcall=True):
        """
        Disconnect from an APRS interface.  This removes the hook to the
        received_msg signal and removes that APRS interface's call-sign/SSID
        from the "mydigi" list.
        """
        if rmcall:
            self.rmaliases(aprsint.mycall)
        aprsint.received_msg.disconnect(self._on_receive)

    def _on_receive(self, interface, frame, **kwargs):
        """
        Handle the incoming to-be-digipeated message.
        """
        # First, have we already digipeated this?
        self._log.debug('On receive call-back: interface=%s, frame=%s',
                interface, frame)
        mycall = interface.mycall
        idx = None
        alias = None
        rem_hops = None

        prev = None
        for (digi_idx, digi) in enumerate(frame.header.repeaters):
            if digi.normalised in self._mydigi:
                self._log.debug('MYDIGI digipeat for %s, last was %s',
                        digi, prev)
                if ((prev is None) or prev.ch) and (not digi.ch):
                    # This is meant to be directly digipeated by us!
                    self._on_transmit(
                            interface=interface,
                            alias=alias,
                            frame=frame.copy(
                                header=AX25FrameHeader(
                                    destination=frame.header.destination,
                                    source=frame.header.source,
                                    repeaters=frame.header.repeaters.replace(
                                        alias=digi,
                                        address=mycall.copy(ch=True)
                                    ),
                                    cr=frame.header.cr
                                )
                            )
                    )
                return
            else:
                # Is this a WIDEn/TRACEn call?
                match = DIGI_RE.match(digi.callsign)
                self._log.debug('WIDEn-N?  digi=%s match=%s', digi, match)
                if match:
                    # It is
                    idx = digi_idx
                    alias = digi
                    rem_hops = min(digi.ssid, int(match.group(2)))
                    break
                else:
                    prev = digi

        if alias is None:
            # The path did not mention a WIDEn digi call
            self._log.debug('No alias, ignoring frame')
            return

        if rem_hops == 0:
            # Number of hops expired, do not digipeat this
            self._log.debug('Hops exhausted, ignoring frame')
            return

        # This is to be digipeated.
        digi_path = list(frame.header.repeaters[:idx]) \
                + [mycall.copy(ch=True)]
        if rem_hops > 1:
            # There are more hops left, tack the next hop on
            digi_path.append(alias.copy(
                    ssid=rem_hops - 1,
                    ch=False
            ))
        digi_path.extend(frame.header.repeaters[idx+1:])

        self._on_transmit(
                interface=interface,
                alias=alias,
                frame=frame.copy(
                header=AX25FrameHeader(
                    destination=frame.header.destination,
                    source=frame.header.source,
                    repeaters=digi_path,
                    cr=frame.header.cr
                )
            )
        )

    def _on_transmit(self, interface, alias, frame):
        """
        Transmit a message to be digipeated.  This function is a wrapper
        around the interface.transmit method so we can support subclasses
        that do more advanced routing such as cross-interface digipeating.
        """
        interface.transmit(frame)



#!/usr/bin/env python3

"""
Tests for signalslot wrappers
"""

from nose.tools import eq_, assert_set_equal

from aioax25.signal import Signal, Slot, OneshotSlot

def test_connect():
    """
    Test connect links a slot function to the signal.
    """
    calls = []
    signal = Signal()
    signal.connect(lambda **kw : calls.append(kw))
    signal.emit(myarg=123)
    signal.emit(myarg=456)

    eq_(len(calls),2)

    # First call should have myarg=123
    call = calls.pop(0)
    assert_set_equal(set(call.keys()), set(['myarg']))
    eq_(call['myarg'], 123)

    # Last call should have myarg=456
    call = calls.pop(0)
    assert_set_equal(set(call.keys()), set(['myarg']))
    eq_(call['myarg'], 456)

def test_connect_oneshot():
    """
    Test connect_oneshot links a slot function to the signal in one-shot mode.
    """
    calls = []
    signal = Signal()
    signal.connect_oneshot(lambda **kw : calls.append(kw))
    signal.emit(myarg=123)
    signal.emit(myarg=456)

    eq_(len(calls),1)

    # Only call should have myarg=123
    call = calls.pop(0)
    assert_set_equal(set(call.keys()), set(['myarg']))
    eq_(call['myarg'], 123)

def test_find_slot():
    """
    Test _find_slot can locate a slot
    """
    slot_fn = lambda **kw : None
    signal = Signal()
    signal.connect(slot_fn)

    slot = signal._find_slot(slot_fn)
    assert isinstance(slot, Slot)
    assert not isinstance(slot, OneshotSlot)

def test_find_oneshot_slot():
    """
    Test _find_slot can locate a one-shot slot
    """
    slot_fn = lambda **kw : None
    signal = Signal()
    signal.connect_oneshot(slot_fn)

    slot = signal._find_slot(slot_fn)
    assert isinstance(slot, Slot)   # Subclass
    assert isinstance(slot, OneshotSlot)

def test_disconnect():
    """
    Test disconnect detaches a slot function from the signal.
    """
    calls = []
    slot_fn = lambda **kw : calls.append(kw)
    signal = Signal()
    signal.connect(slot_fn)
    signal.emit(myarg=123)
    signal.disconnect(slot_fn)
    signal.emit(myarg=456)

    eq_(len(calls),1)

    # Only call should have myarg=123
    call = calls.pop(0)
    assert_set_equal(set(call.keys()), set(['myarg']))
    eq_(call['myarg'], 123)

def test_is_connected():
    """
    Test is_connected returns True for connected signals.
    """
    slot_fn = lambda **kw : None
    signal = Signal()
    signal.connect(slot_fn)
    assert signal.is_connected(slot_fn)



#!/usr/bin/env python3

"""
APRS framing.
"""

from ..frame import AX25UnnumberedInformationFrame
from .datatype import APRSDataType


class APRSFrame(AX25UnnumberedInformationFrame):
    """
    This is a helper sub-class for encoding and decoding APRS messages into
    AX.25 frames.
    """

    DATA_TYPE_HANDLERS = {}

    @classmethod
    def decode(cls, uiframe, log):
        """
        Decode the given UI frame (AX25UnnumberedInformationFrame) to a
        suitable APRSFrame sub-class.
        """
        # Do not decode if not the APRS PID value
        if uiframe.pid != cls.PID_NO_L3:
            # Clearly not an APRS message
            log.debug('Frame has wrong PID for APRS')
            return uiframe

        if len(uiframe.payload) == 0:
            log.debug('Frame has no payload data')
            return uiframe

        try:
            # Inspect the first byte.
            type_code = APRSDataType(uiframe.payload[0])
            handler_class = cls.DATA_TYPE_HANDLERS[type_code]

            # Decode the payload as text
            payload = uiframe.payload.decode('US-ASCII')

            return handler_class.decode(uiframe, payload, log)
        except:
            # Not decodable, leave as-is
            log.debug('Failed to decode as APRS', exc_info=1)
            return uiframe

    def __init__(self, destination, source, payload, repeaters=None,
            pf=False, cr=False):
        super(APRSFrame, self).__init__(
                destination=destination,
                source=source,
                pid=self.PID_NO_L3, # APRS spec
                payload=payload,
                repeaters=repeaters,
                pf=pf, cr=cr)



#!/usr/bin/env python3

from aioax25.frame import AX25Address, AX25Path
from nose.tools import eq_


def test_decode():
    """
    Test given a list of strings, AX25Path decodes them.
    """
    path = AX25Path('VK4MSL', 'VK4RZB', 'VK4RZA')
    eq_(path._path[0]._callsign, 'VK4MSL')
    eq_(path._path[1]._callsign, 'VK4RZB')
    eq_(path._path[2]._callsign, 'VK4RZA')

def test_addresses():
    """
    Test given a list of AX25Addresses, AX25Path passes them in.
    """
    path = AX25Path(
            AX25Address.decode('VK4MSL'),
            AX25Address.decode('VK4RZB'),
            AX25Address.decode('VK4RZA')
    )
    eq_(path._path[0]._callsign, 'VK4MSL')
    eq_(path._path[1]._callsign, 'VK4RZB')
    eq_(path._path[2]._callsign, 'VK4RZA')

def test_str():
    """
    Test we can return the canonical format for a repeater path.
    """
    path = AX25Path('VK4MSL', 'VK4RZB', 'VK4RZA')
    eq_(str(path), 'VK4MSL,VK4RZB,VK4RZA')

def test_repr():
    """
    Test we can return the Python representation for a repeater path.
    """
    path = AX25Path('VK4MSL', 'VK4RZB', 'VK4RZA')
    eq_(repr(path), ('AX25Path('\
                        'AX25Address('\
                            'callsign=VK4MSL, ssid=0, ch=False, '\
                            'res0=True, res1=True, extension=False'\
                        '), '\
                        'AX25Address('\
                            'callsign=VK4RZB, ssid=0, ch=False, '\
                            'res0=True, res1=True, extension=False'\
                        '), '\
                        'AX25Address('\
                            'callsign=VK4RZA, ssid=0, ch=False, '\
                            'res0=True, res1=True, extension=False'\
                        ')'\
                    ')'))

def test_reply():
    """
    Test 'reply' returns a reverse path for sending replies.
    """
    path = AX25Path('VK4MSL*', 'VK4RZB*', 'VK4RZA*')
    reply = path.reply
    eq_(str(reply), 'VK4RZA,VK4RZB,VK4MSL')

def test_reply_unused():
    """
    Test 'reply' ignores unused repeaters.
    """
    path = AX25Path('VK4MSL*', 'VK4RZB*', 'VK4RZA')
    reply = path.reply
    eq_(str(reply), 'VK4RZB,VK4MSL')

def test_replace():
    """
    Test we can replace an alias with our own call-sign.
    """
    path1 = AX25Path('WIDE2-2', 'WIDE1-1')
    path2 = path1.replace('WIDE2-2', 'VK4MSL*')
    eq_(str(path2), 'VK4MSL*,WIDE1-1')



#!/usr/bin/env python3

from aioax25.frame import AX25Frame, AX25RawFrame, \
        AX25UnnumberedInformationFrame, AX25FrameRejectFrame, \
        AX25UnnumberedFrame
from nose.tools import eq_
from ..hex import from_hex, hex_cmp


def test_decode_incomplete():
    """
    Test that an incomplete frame does not cause a crash.
    """
    try:
        AX25Frame.decode(
                from_hex(
                    'ac 96 68 84 ae 92 e0'      # Destination
                    'ac 96 68 9a a6 98 61'      # Source
                )
        )
        assert False, 'This should not have worked'
    except ValueError as e:
        eq_(str(e), 'Insufficient packet data')

def test_decode_iframe():
    """
    Test that an I-frame gets decoded to a raw frame.
    """
    frame = AX25Frame.decode(
            from_hex(
                'ac 96 68 84 ae 92 e0'      # Destination
                'ac 96 68 9a a6 98 61'      # Source
                '00'                        # Control byte
                '11 22 33 44 55 66 77'      # Payload
            )
    )
    assert isinstance(frame, AX25RawFrame), 'Did not decode to raw frame'
    eq_(frame.control, 0x00)
    hex_cmp(frame.frame_payload, '11 22 33 44 55 66 77')

def test_decode_sframe():
    """
    Test that an S-frame gets decoded to a raw frame.
    """
    frame = AX25Frame.decode(
            from_hex(
                'ac 96 68 84 ae 92 e0'      # Destination
                'ac 96 68 9a a6 98 61'      # Source
                '01'                        # Control byte
                '11 22 33 44 55 66 77'      # Payload
            )
    )
    assert isinstance(frame, AX25RawFrame), 'Did not decode to raw frame'
    eq_(frame.control, 0x01)
    hex_cmp(frame.frame_payload, '11 22 33 44 55 66 77')

def test_decode_uframe():
    """
    Test that a U-frame gets decoded to an unnumbered frame.
    """
    frame = AX25Frame.decode(
            from_hex(
                'ac 96 68 84 ae 92 e0'      # Destination
                'ac 96 68 9a a6 98 61'      # Source
                'c3'                        # Control byte
            )
    )
    assert isinstance(frame, AX25UnnumberedFrame), \
            'Did not decode to unnumbered frame'
    eq_(frame.modifier, 0xc3)
    hex_cmp(frame.frame_payload, '')

def test_decode_uframe_payload():
    """
    Test that U-frames other than FRMR and UI are forbidden to have payloads.
    """
    try:
        AX25Frame.decode(
                from_hex(
                    'ac 96 68 84 ae 92 e0'      # Destination
                    'ac 96 68 9a a6 98 61'      # Source
                    'c3 11 22 33'               # Control byte
                )
        )
        assert False, 'Should not have worked'
    except ValueError as e:
        eq_(str(e), 'Unnumbered frames (other than UI and '\
                            'FRMR) do not have payloads')

def test_decode_frmr():
    """
    Test that a FRMR gets decoded to a frame reject frame.
    """
    frame = AX25Frame.decode(
            from_hex(
                'ac 96 68 84 ae 92 e0'      # Destination
                'ac 96 68 9a a6 98 61'      # Source
                '87'                        # Control byte
                '11 22 33'                  # Payload
            )
    )
    assert isinstance(frame, AX25FrameRejectFrame), \
            'Did not decode to FRMR frame'
    eq_(frame.modifier, 0x87)
    eq_(frame.w, True)
    eq_(frame.x, False)
    eq_(frame.y, False)
    eq_(frame.z, False)
    eq_(frame.vr, 1)
    eq_(frame.frmr_cr, False)
    eq_(frame.vs, 1)

def test_decode_frmr_len():
    """
    Test that a FRMR must have 3 byte payload.
    """
    try:
        AX25Frame.decode(
                from_hex(
                    'ac 96 68 84 ae 92 e0'      # Destination
                    'ac 96 68 9a a6 98 61'      # Source
                    '87'                        # Control byte
                    '11 22'                     # Payload
                )
        )
        assert False, 'Should not have worked'
    except ValueError as e:
        eq_(str(e), 'Payload of FRMR must be 3 bytes')

def test_decode_ui():
    """
    Test that a UI gets decoded to an unnumbered information frame.
    """
    frame = AX25Frame.decode(
            from_hex(
                'ac 96 68 84 ae 92 e0'      # Destination
                'ac 96 68 9a a6 98 61'      # Source
                '03 11 22 33'               # Control byte
            )
    )
    assert isinstance(frame, AX25UnnumberedInformationFrame), \
            'Did not decode to UI frame'
    eq_(frame.pid, 0x11)
    hex_cmp(frame.payload, '22 33')

def test_decode_ui_len():
    """
    Test that a UI must have at least one byte payload.
    """
    try:
        AX25Frame.decode(
                from_hex(
                    'ac 96 68 84 ae 92 e0'      # Destination
                    'ac 96 68 9a a6 98 61'      # Source
                    '03'                        # Control byte
                )
        )
        assert False, 'Should not have worked'
    except ValueError as e:
        eq_(str(e), 'Payload of UI must be at least one byte')

def test_encode_raw():
    """
    Test that we can encode a raw frame.
    """
    # Yes, this is really a UI frame.
    frame = AX25RawFrame(
            destination='VK4BWI',
            source='VK4MSL',
            cr=True,
            control=0x03,
            payload=b'\xf0This is a test'
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 e0'                          # Destination
            'ac 96 68 9a a6 98 61'                          # Source
            '03'                                            # Control
            'f0 54 68 69 73 20 69 73 20 61 20 74 65 73 74'  # Payload
    )

def test_encode_uframe():
    """
    Test that we can encode a U-frame.
    """
    frame = AX25UnnumberedFrame(
            destination='VK4BWI',
            source='VK4MSL',
            modifier=0xe7,
            cr=True
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 e0'                          # Destination
            'ac 96 68 9a a6 98 61'                          # Source
            'e7'                                            # Control
    )

def test_encode_frmr():
    """
    Test that we can encode a FRMR.
    """
    frame = AX25FrameRejectFrame(
            destination='VK4BWI',
            source='VK4MSL',
            w=True, x=False, y=True, z=False,
            vr=1, frmr_cr=False, vs=2, frmr_control=0xaa,
            cr=True
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 e0'                          # Destination
            'ac 96 68 9a a6 98 61'                          # Source
            '87'                                            # Control
            '05'                                            # W/X/Y/Z
            '24'                                            # VR/CR/VS
            'aa'                                            # FRMR Control
    )

def test_encode_ui():
    """
    Test that we can encode a UI frame.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI',
            source='VK4MSL',
            cr=True,
            pid=0xf0,
            payload=b'This is a test'
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 e0'                          # Destination
            'ac 96 68 9a a6 98 61'                          # Source
            '03'                                            # Control
            'f0'                                            # PID
            '54 68 69 73 20 69 73 20 61 20 74 65 73 74'     # Payload
    )

def test_encode_pf():
    """
    Test we can set the PF bit on a frame.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI',
            source='VK4MSL',
            cr=True, pf=True,
            pid=0xf0,
            payload=b'This is a test'
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 e0'                          # Destination
            'ac 96 68 9a a6 98 61'                          # Source
            '13'                                            # Control
            'f0'                                            # PID
            '54 68 69 73 20 69 73 20 61 20 74 65 73 74'     # Payload
    )

def test_encode_frmr_w():
    """
    Test we can set the W bit on a FRMR frame.
    """
    frame = AX25FrameRejectFrame(
            destination='VK4BWI',
            source='VK4MSL',
            w=True, x=False, y=False, z=False,
            vr=0, vs=0, frmr_control=0, frmr_cr=False
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            '87'                                            # Control
            '01 00 00'                                      # FRMR data
    )

def test_encode_frmr_x():
    """
    Test we can set the X bit on a FRMR frame.
    """
    frame = AX25FrameRejectFrame(
            destination='VK4BWI',
            source='VK4MSL',
            w=False, x=True, y=False, z=False,
            vr=0, vs=0, frmr_control=0, frmr_cr=False
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            '87'                                            # Control
            '02 00 00'                                      # FRMR data
    )

def test_encode_frmr_y():
    """
    Test we can set the Y bit on a FRMR frame.
    """
    frame = AX25FrameRejectFrame(
            destination='VK4BWI',
            source='VK4MSL',
            w=False, x=False, y=True, z=False,
            vr=0, vs=0, frmr_control=0, frmr_cr=False
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            '87'                                            # Control
            '04 00 00'                                      # FRMR data
    )

def test_encode_frmr_z():
    """
    Test we can set the Z bit on a FRMR frame.
    """
    frame = AX25FrameRejectFrame(
            destination='VK4BWI',
            source='VK4MSL',
            w=False, x=False, y=False, z=True,
            vr=0, vs=0, frmr_control=0, frmr_cr=False
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            '87'                                            # Control
            '08 00 00'                                      # FRMR data
    )

def test_encode_frmr_cr():
    """
    Test we can set the CR bit on a FRMR frame.
    """
    frame = AX25FrameRejectFrame(
            destination='VK4BWI',
            source='VK4MSL',
            w=False, x=False, y=False, z=False,
            vr=0, vs=0, frmr_control=0, frmr_cr=True
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            '87'                                            # Control
            '00 10 00'                                      # FRMR data
    )

def test_encode_frmr_vr():
    """
    Test we can set the V(R) field on a FRMR frame.
    """
    frame = AX25FrameRejectFrame(
            destination='VK4BWI',
            source='VK4MSL',
            w=False, x=False, y=False, z=False,
            vr=5, vs=0, frmr_control=0, frmr_cr=False
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            '87'                                            # Control
            '00 a0 00'                                      # FRMR data
    )

def test_encode_frmr_vs():
    """
    Test we can set the V(S) field on a FRMR frame.
    """
    frame = AX25FrameRejectFrame(
            destination='VK4BWI',
            source='VK4MSL',
            w=False, x=False, y=False, z=False,
            vr=0, vs=5, frmr_control=0, frmr_cr=False
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            '87'                                            # Control
            '00 0a 00'                                      # FRMR data
    )

def test_encode_frmr_frmr_ctrl():
    """
    Test we can set the FRMR Control field on a FRMR frame.
    """
    frame = AX25FrameRejectFrame(
            destination='VK4BWI',
            source='VK4MSL',
            w=False, x=False, y=False, z=False,
            vr=0, vs=0, frmr_control=0x55, frmr_cr=False
    )
    hex_cmp(bytes(frame),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            '87'                                            # Control
            '00 00 55'                                      # FRMR data
    )

def test_raw_copy():
    """
    Test we can make a copy of a raw frame.
    """
    frame = AX25RawFrame(
            destination='VK4BWI',
            source='VK4MSL',
            control=0xab,
            payload=b'This is a test'
    )
    framecopy = frame.copy()
    assert framecopy is not frame

    hex_cmp(bytes(framecopy),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            'ab'                                            # Control
            '54 68 69 73 20 69 73 20 61 20 74 65 73 74'     # Payload
    )

def test_u_copy():
    """
    Test we can make a copy of a unnumbered frame.
    """
    frame = AX25UnnumberedFrame(
            destination='VK4BWI',
            source='VK4MSL',
            modifier=0x43    # Disconnect
    )
    framecopy = frame.copy()
    assert framecopy is not frame

    hex_cmp(bytes(framecopy),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            '43'                                            # Control
    )

def test_ui_copy():
    """
    Test we can make a copy of a unnumbered information frame.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI',
            source='VK4MSL',
            cr=True,
            pid=0xf0,
            payload=b'This is a test'
    )
    framecopy = frame.copy()
    assert framecopy is not frame

    hex_cmp(bytes(framecopy),
            'ac 96 68 84 ae 92 e0'                          # Destination
            'ac 96 68 9a a6 98 61'                          # Source
            '03'                                            # Control
            'f0'                                            # PID
            '54 68 69 73 20 69 73 20 61 20 74 65 73 74'     # Payload
    )

def test_frmr_copy():
    """
    Test we can copy a FRMR frame.
    """
    frame = AX25FrameRejectFrame(
            destination='VK4BWI',
            source='VK4MSL',
            w=False, x=False, y=False, z=False,
            vr=0, vs=0, frmr_control=0x55, frmr_cr=False
    )
    framecopy = frame.copy()

    assert framecopy is not frame
    hex_cmp(bytes(framecopy),
            'ac 96 68 84 ae 92 60'                          # Destination
            'ac 96 68 9a a6 98 e1'                          # Source
            '87'                                            # Control
            '00 00 55'                                      # FRMR data
    )

def test_raw_str():
    """
    Test we can get a string representation of a raw frame.
    """
    frame = AX25RawFrame(
            destination='VK4BWI',
            source='VK4MSL',
            control=0xab,
            payload=b'This is a test'
    )
    eq_(str(frame), "VK4MSL>VK4BWI")

def test_ui_str():
    """
    Test we can get a string representation of a UI frame.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='VK4BWI',
            source='VK4MSL',
            cr=True,
            pid=0xf0,
            payload=b'This is a test'
    )
    eq_(str(frame), "VK4MSL>VK4BWI: PID=0xf0 Payload=b'This is a test'")



#!/usr/bin/env python3

import logging

from nose.tools import eq_, assert_is, assert_is_not

from aioax25.aprs.frame import APRSFrame
from aioax25.aprs.message import \
        APRSMessageAckFrame, APRSMessageRejFrame, APRSMessageFrame
from aioax25.frame import AX25UnnumberedInformationFrame

from ..loop import DummyLoop

def test_decode_wrong_pid():
    """
    Test the decode routine ignores UI frames with wrong PID.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0x80,   # Not correct for APRS, should not get decoded.
            payload=b'Test frame that\'s not APRS'
    )
    decoded = APRSFrame.decode(frame, logging.getLogger('decoder'))
    assert_is(decoded, frame)

def test_decode_no_payload():
    """
    Test the decode routine rejects frames without a payload.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'' # Empty payload
    )
    decoded = APRSFrame.decode(frame, logging.getLogger('decoder'))
    assert_is(decoded, frame)

def test_decode_unknown_type():
    """
    Test the decode routine rejects frames it cannot recognise.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b'X A mystery frame X'
    )
    decoded = APRSFrame.decode(frame, logging.getLogger('decoder'))
    assert_is(decoded, frame)

def test_decode_message():
    """
    Test the decode routine can recognise a message frame.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b':VK4MDL-7 :Hi'
    )
    decoded = APRSFrame.decode(frame, logging.getLogger('decoder'))
    assert_is_not(decoded, frame)
    assert isinstance(decoded, APRSMessageFrame)

def test_decode_message_confirmable():
    """
    Test the decode routine can recognise a confirmable message frame.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b':VK4MDL-7 :Hi{14'
    )
    decoded = APRSFrame.decode(frame, logging.getLogger('decoder'))
    assert_is_not(decoded, frame)
    assert isinstance(decoded, APRSMessageFrame)
    eq_(decoded.msgid, '14')

def test_decode_message_ack():
    """
    Test the decode routine can recognise a message acknowledgement frame.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b':VK4MDL-7 :ack2'
    )
    decoded = APRSFrame.decode(frame, logging.getLogger('decoder'))
    assert_is_not(decoded, frame)
    assert isinstance(decoded, APRSMessageAckFrame)
    eq_(decoded.msgid, '2')

def test_decode_message_rej():
    """
    Test the decode routine can recognise a message rejection frame.
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b':VK4MDL-7 :rej3'
    )
    decoded = APRSFrame.decode(frame, logging.getLogger('decoder'))
    assert_is_not(decoded, frame)
    assert isinstance(decoded, APRSMessageRejFrame)
    eq_(decoded.msgid, '3')



#!/usr/bin/env python3

"""
Convenience wrappers around `signalslot`
"""

from signalslot import Signal as BaseSignal, Slot as BaseSlot
from weakref import ref
import logging


class Slot(BaseSlot):
    """
    Wrapper class around a slot function.  This will wrap the given slot up in
    an exception handler.  It is the caller's responsibility to ensure the
    slot function handles its own exceptions.
    """
    def __init__(self, slot_fn, **kwargs):
        super(Slot, self).__init__(slot_fn)
        self._slot_kwargs = kwargs

    def __call__(self, **kwargs):
        try:
            call_kwargs = self._slot_kwargs.copy()
            call_kwargs.update(kwargs)
            super(Slot, self).__call__(**call_kwargs)
        except:
            logging.getLogger(self.__class__.__module__).exception(
                    'Exception in slot %s', self.func
            )


class OneshotSlot(Slot):
    """
    Helper class that calls a slot exactly once.
    """
    def __init__(self, signal, slot_fn, **kwargs):
        self._signal = ref(signal)
        super(OneshotSlot, self).__init__(slot_fn, **kwargs)

    def __call__(self, **kwargs):
        super(OneshotSlot, self).__call__(**kwargs)
        signal = self._signal()
        if signal:
            signal.disconnect(self)


class Signal(BaseSignal):
    """
    Wrap the `signalslot.Signal` so that *all* "slots" get called, regardless
    of whether they return something or not, or throw exceptions.
    """
    def connect(self, slot, **kwargs):
        """
        Connect a slot to the signal.  This will wrap the given slot up in
        an exception handler.  It is the caller's responsibility to ensure
        the slot handles its own exceptions.
        """
        super(Signal, self).connect(Slot(slot, **kwargs))

    def connect_oneshot(self, slot, **kwargs):
        """
        Connect a slot to the signal, and call it exactly once when the
        signal fires.  (Disconnect after calling.)
        """
        super(Signal, self).connect(OneshotSlot(self, slot, **kwargs))

    def _find_slot(self, slot):
        """
        Locate a slot connected to the signal.
        """
        for maybe_slot in super(Signal, self).slots:
            if isinstance(maybe_slot, Slot) and \
                    (maybe_slot.func is slot):
                # Here it is
                return maybe_slot
            elif maybe_slot is slot:
                # Here it is
                return maybe_slot

    def disconnect(self, slot):
        """
        Disconnect the first located matching slot.
        """
        slot = self._find_slot(slot)
        if slot:
            super(Signal, self).disconnect(slot)

    def is_connected(self, slot):
        """
        Check if a callback slot is connected to this signal.
        """
        if isinstance(slot, Slot):
            return super(Signal, self).is_connected(slot.func)
        else:
            return super(Signal, self).is_connected(slot)



#!/usr/bin/env python3

"""
APRS messaging handler.
"""

import re
import weakref
import random
from ..signal import Signal
from enum import Enum

from ..frame import AX25Address
from .frame import APRSFrame
from .datatype import APRSDataType


class APRSMessageHandler(object):
    """
    The APRS message handler is a helper class that handles the
    retransmissions, timeouts and responses to an APRS message.
    """
    class HandlerState(Enum):
        INIT    = 0
        SEND    = 1
        RETRY   = 2
        SUCCESS = 3
        REJECT  = -1
        CANCEL  = -2
        TIMEOUT = -3
        FAIL    = -4

    def __init__(self, aprshandler, addressee, path, message, log):
        self._log = log
        # Initialise our timer and retry counter
        self._timeout_duration \
                = aprshandler._retransmit_timeout_base \
                + (random.random() * aprshandler._retransmit_timeout_rand)
        self._retransmit_count = aprshandler._retransmit_count
        self._retransmit_timeout_scale = aprshandler._retransmit_timeout_scale
        self._loop = aprshandler._loop
        self._tx_frame = APRSMessageFrame(
                destination=addressee,
                source=aprshandler.mycall,
                addressee=addressee,
                message=message,
                msgid=aprshandler._next_msgid,
                repeaters=[
                    AX25Address.decode(call).normalised for call in path
                ]
        )

        self._aprshandler = weakref.ref(aprshandler)
        self._retransmit_timeout = None
        self._response = None

        self.done = Signal()
        self._state = self.HandlerState.INIT
        self._log.debug('Initialised handler for %s state %s',
                self.msgid, self.state)

    @property
    def frame(self):
        return self._tx_frame

    @property
    def addressee(self):
        return self._tx_frame.addressee

    @property
    def msgid(self):
        return self._tx_frame.msgid

    @property
    def state(self):
        return self._state

    @property
    def response(self):
        return self._response

    def cancel(self):
        self._stop_timer()
        self._enter_state(self.HandlerState.CANCEL)

    def _send(self):
        # Stop any timers
        self._stop_timer()

        self._log.info('Preparing to send %s (state %s)',
                self.msgid, self.state)

        # What state are we in?
        if self.state == self.HandlerState.INIT:
            next_state = self.HandlerState.SEND
        elif self.state in (self.HandlerState.SEND, self.HandlerState.RETRY):
            # Have we exhausted the retries?
            if self._retransmit_count <= 0:
                self._enter_state(self.HandlerState.TIMEOUT)
                return
            self._retransmit_count -= 1
            next_state = self.HandlerState.RETRY
        else:
            self._log.warning('Attempt to send %s in state %s',
                    self.msgid, self.state)
            raise RuntimeError('Incorrect state %s' % self.state)

        handler = self._aprshandler()
        if handler is None:
            # No handler, so can't send
            self._enter_state(self.HandlerState.FAIL)
            return

        # Set the time-out timer
        self._retransmit_timeout = self._loop.call_later(
                self._timeout_duration,
                self._on_timeout
        )
        self._timeout_duration *= self._retransmit_timeout_scale

        # Send the frame
        handler.transmit(self.frame)
        self._enter_state(next_state)

    def _stop_timer(self):
        self._log.debug('Cancelling timer for %s', self.msgid)
        if self._retransmit_timeout is not None:
            self._retransmit_timeout.cancel()
            self._retransmit_timeout = None

    def _on_timeout(self):
        self._log.warning('Time-out waiting for reponse to %s', self.frame)
        self._loop.call_soon(self._send)

    def _on_response(self, response):
        self._stop_timer()
        self._log.info('%s Received response %s to frame %s',
                self.msgid, response, self.frame)
        if self.state not in (self.HandlerState.SEND, self.HandlerState.RETRY):
            # Ignore the message, we are no longer interested
            return

        self._response = response
        if isinstance(response, APRSMessageAckFrame):
            self._enter_state(self.HandlerState.SUCCESS)
        else:
            self._enter_state(self.HandlerState.REJECT)

    def _enter_state(self, state):
        self._log.debug('%s entering state %s', self.msgid, state)
        self._state = state
        if state in (self.HandlerState.SUCCESS,
                    self.HandlerState.REJECT,
                    self.HandlerState.TIMEOUT,
                    self.HandlerState.CANCEL):
            self._log.info('%s is done', self.frame)
            # These are final states.
            handler = self._aprshandler()
            if handler:
                handler._on_msg_handler_finish(self.msgid)
            self.done.emit(handler=self, state=state)


class APRSMessageFrame(APRSFrame):

    MSGID_RE = re.compile(r'{([0-9A-Za-z]+)(\r?)$')
    ACKREJ_RE = re.compile(r'^(ack|rej)([0-9A-Za-z]+)$')

    @classmethod
    def decode(cls, uiframe, payload, log):
        if (payload[0] != ':') and (payload[10] != ':'):
            raise ValueError('Not a message frame: %r' % payload)

        addressee = AX25Address.decode(payload[1:10].strip())
        message = payload[11:]

        match = cls.ACKREJ_RE.match(message)
        if match:
            ackrej = match.group(1)
            msgid = match.group(2)
            if ackrej == 'ack':
                # This is an ACK
                return APRSMessageAckFrame(
                    destination=uiframe.header.destination,
                    source=uiframe.header.source,
                    addressee=addressee,
                    msgid=msgid,
                    repeaters=uiframe.header.repeaters,
                    pf=uiframe.pf, cr=uiframe.header.cr
                )
            else:
                # Must be a rejection then
                return APRSMessageRejFrame(
                    destination=uiframe.header.destination,
                    source=uiframe.header.source,
                    addressee=addressee,
                    msgid=msgid,
                    repeaters=uiframe.header.repeaters,
                    pf=uiframe.pf, cr=uiframe.header.cr
                )

        match = cls.MSGID_RE.search(message)
        if match:
            msgid = match.group(1)
            message = message[:-(len(msgid)+1)]
        else:
            msgid = None

        return cls(
                destination=uiframe.header.destination,
                source=uiframe.header.source,
                addressee=addressee,
                message=message,
                msgid=msgid,
                repeaters=uiframe.header.repeaters,
                pf=uiframe.pf, cr=uiframe.header.cr
        )

    def __init__(self, destination, source, addressee, message,
            msgid=None, repeaters=None, pf=False, cr=False):

        self._addressee = AX25Address.decode(addressee).normalised
        self._msgid = msgid
        self._message = message

        payload = ':%-9s:%s' % (
            self._addressee,
            message[0:67]
        )

        if msgid is not None:
            msgid = str(msgid)
            if len(msgid) > 5:
                raise ValueError('message ID %r too long' % msgid)
            payload += '{%s' % msgid

        super(APRSMessageFrame, self).__init__(
                destination=destination,
                source=source,
                payload=payload.encode('US-ASCII'),
                repeaters=repeaters, pf=pf, cr=cr)

    @property
    def addressee(self):
        return self._addressee

    @property
    def msgid(self):
        return self._msgid

    @property
    def message(self):
        return self._message

    def _copy(self):
        return self.__class__(
                destination=self.header.destination,
                source=self.header.source,
                repeaters=self.header.repeaters,
                cr=self.header.cr,
                pf=self.pf,
                addressee=self.addressee,
                msgid=self.msgid,
                message=self.message
        )

APRSFrame.DATA_TYPE_HANDLERS[APRSDataType.MESSAGE] = APRSMessageFrame


class APRSMessageAckFrame(APRSMessageFrame):
    def __init__(self, destination, source, addressee, msgid,
            repeaters=None, pf=False, cr=False):
        super(APRSMessageAckFrame, self).__init__(
            destination=destination,
            source=source,
            addressee=addressee,
            message='ack%s' % msgid,
            msgid=None, # Don't encode the message ID a second time
            repeaters=repeaters, pf=pf, cr=cr)

        self._msgid = msgid

    def _copy(self):
        return self.__class__(
                destination=self.header.destination,
                source=self.header.source,
                repeaters=self.header.repeaters,
                cr=self.header.cr,
                pf=self.pf,
                addressee=self.addressee,
                msgid=self.msgid
        )


class APRSMessageRejFrame(APRSMessageFrame):
    def __init__(self, destination, source, addressee, msgid,
            repeaters=None, pf=False, cr=False):
        super(APRSMessageRejFrame, self).__init__(
            destination=destination,
            source=source,
            addressee=addressee,
            message='rej%s' % msgid,
            msgid=None, # Don't encode the message ID a second time
            repeaters=repeaters, pf=pf, cr=cr)

        self._msgid = msgid

    def _copy(self):
        return self.__class__(
                destination=self.header.destination,
                source=self.header.source,
                repeaters=self.header.repeaters,
                cr=self.header.cr,
                pf=self.pf,
                addressee=self.addressee,
                msgid=self.msgid
        )



#!/usr/bin/env python3

from nose.tools import eq_, assert_set_equal, assert_is, assert_greater, \
        assert_less

import logging
from functools import partial
from signalslot import Signal

from aioax25.aprs import APRSInterface
from aioax25.frame import AX25UnnumberedInformationFrame
from aioax25.aprs.message import APRSMessageFrame, APRSMessageHandler

from ..loop import DummyLoop


class DummyAX25Interface(object):
    def __init__(self):
        self._loop = DummyLoop()
        self.bind_calls = []
        self.transmitted = []
        self.received_msg = Signal()

    def bind(self, callback, callsign, ssid=0, regex=False):
        self.bind_calls.append((callback, callsign, ssid, regex))

    def transmit(self, frame):
        self.transmitted.append(frame)


class DummyMessageHandler(object):
    def _on_response(self):
        pass


def test_constructor_log():
    """
    Test the constructor can accept a logger instance.
    """
    ax25int = DummyAX25Interface()
    log = logging.getLogger('aprslog')
    aprsint = APRSInterface(ax25int, 'VK4MSL-10', log=log)
    assert log is aprsint._log

def test_transmit_exception():
    """
    Test that transmit swallows exceptions.
    """
    ax25int = DummyAX25Interface()

    # Stub the transmit so it fails
    calls = []
    def stub(*args):
        calls.append(args)
        raise RuntimeError('Oopsie')
    ax25int.transmit = stub

    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    aprsint.transmit(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123
            )
    )

    # Transmit should have been called
    eq_(len(calls), 1)

def test_send_message_oneshot():
    """
    Test that send_message in one-shot mode generates a message frame.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    res = aprsint.send_message(
            'VK4MDL-7', 'Hi', oneshot=True
    )

    # We don't get a return value
    assert_is(res, None)

    # No message handler should be registered with the interface
    eq_(len(aprsint._pending_msg), 0)

    # The frame is passed to the AX.25 interface
    eq_(len(ax25int.transmitted), 1)
    frame = ax25int.transmitted.pop(0)

    # Frame is a APRS message frame
    assert isinstance(frame, APRSMessageFrame)

    # There is no pending messages
    eq_(len(aprsint._pending_msg), 0)

def test_send_message_confirmable():
    """
    Test that send_message in confirmable mode generates a message handler.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    res = aprsint.send_message(
            'VK4MDL-7', 'Hi', oneshot=False
    )

    # We got back a handler class
    assert isinstance(res, APRSMessageHandler)

    # That message handler should be registered with the interface
    eq_(len(aprsint._pending_msg), 1)
    assert res.msgid in aprsint._pending_msg
    assert_is(aprsint._pending_msg[res.msgid], res)

    # The APRS message handler will have tried sending the message
    eq_(len(ax25int.transmitted), 1)
    frame = ax25int.transmitted.pop(0)

    # Frame is a APRS message frame
    assert isinstance(frame, APRSMessageFrame)

    # Message handler is in 'SEND' state
    eq_(res.state, APRSMessageHandler.HandlerState.SEND)

def test_send_response_oneshot():
    """
    Test that send_response ignores one-shot messages.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    aprsint.send_response(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=None
            )
    )

    # Nothing should be sent
    eq_(len(ax25int.transmitted), 0)

def test_send_response_ack():
    """
    Test that send_response with ack=True sends acknowledgement.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    aprsint.send_response(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123
            ),
            ack=True
    )

    # The APRS message handler will have tried sending the message
    eq_(len(ax25int.transmitted), 1)
    frame = ax25int.transmitted.pop(0)

    # Frame is a APRS message acknowledgement frame
    assert isinstance(frame, APRSMessageFrame)
    eq_(frame.payload, b':VK4MSL-10:ack123')

def test_send_response_rej():
    """
    Test that send_response with ack=False sends rejection.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    aprsint.send_response(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123
            ),
            ack=False
    )

    # The APRS message handler will have tried sending the message
    eq_(len(ax25int.transmitted), 1)
    frame = ax25int.transmitted.pop(0)

    # Frame is a APRS message rejection frame
    assert isinstance(frame, APRSMessageFrame)
    eq_(frame.payload, b':VK4MSL-10:rej123')

def test_hash_frame_mismatch_dest():
    """
    Test that _hash_frame returns different hashes for mismatching destination
    """
    hash1 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    hash2 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-3',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    # These should not be the same
    assert hash1 != hash2

def test_hash_frame_mismatch_src():
    """
    Test that _hash_frame returns different hashes for mismatching source
    """
    hash1 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    hash2 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-11',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    # These should not be the same
    assert hash1 != hash2

def test_hash_frame_mismatch_payload():
    """
    Test that _hash_frame returns different hashes for mismatching source
    """
    hash1 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing 1',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    hash2 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing 2',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    # These should not be the same
    assert hash1 != hash2

def test_hash_frame():
    """
    Test that _hash_frame returns a consistent result regardless of digipeaters
    """
    hash1 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    hash2 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['VK4RZB*','WIDE1-1']
            )
    )

    hash3 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['VK4RZB*','VK4RZA*']
            )
    )

    # These should all be the same
    eq_(hash1, hash2)
    eq_(hash1, hash3)

def test_test_or_add_frame_first():
    """
    Test that _test_or_add_frame returns False for new traffic
    """
    frame = APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
    )
    framedigest = APRSInterface._hash_frame(frame)

    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Try it out
    res = aprsint._test_or_add_frame(frame)

    # We should get 'False' as the response
    eq_(res, False)

    # There should be an entry in our hash table.
    eq_(len(aprsint._msg_expiry), 1)

    # The expiry time should be at least 25 seconds.
    assert_greater(aprsint._msg_expiry.get(framedigest, 0),
            ax25int._loop.time() + 25)

    # A clean-up should have been scheduled.
    eq_(len(ax25int._loop.calls), 1)
    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

def test_test_or_add_frame_repeat():
    """
    Test that _test_or_add_frame returns True for un-expired repeats
    """
    frame = APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
    )
    framedigest = APRSInterface._hash_frame(frame)

    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    # Inject the frame expiry
    expiry_time = aprsint._loop.time() + 1
    aprsint._msg_expiry[framedigest] = expiry_time

    # Try it out
    res = aprsint._test_or_add_frame(frame)

    # We should get 'False' as the response
    eq_(res, True)

    # Expiry should not have changed
    eq_(len(aprsint._msg_expiry), 1)
    eq_(aprsint._msg_expiry[framedigest], expiry_time)

    # Nothing further should be done.
    eq_(len(ax25int._loop.calls), 0)

def test_test_or_add_frame_expired():
    """
    Test that _test_or_add_frame returns False for expired repeats
    """
    frame = APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
    )
    framedigest = APRSInterface._hash_frame(frame)

    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    # Inject the frame expiry
    expiry_time = aprsint._loop.time() - 1
    aprsint._msg_expiry[framedigest] = expiry_time

    # Try it out
    res = aprsint._test_or_add_frame(frame)

    # We should get 'False' as the response
    eq_(res, False)

    # The expiry time should be at least 25 seconds.
    eq_(len(aprsint._msg_expiry), 1)
    assert_greater(aprsint._msg_expiry.get(framedigest, 0),
            ax25int._loop.time() + 25)

    # A clean-up should have been scheduled.
    eq_(len(ax25int._loop.calls), 1)
    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

def test_schedule_dedup_cleanup_no_msg():
    """
    Test _schedule_dedup_cleanup does nothing if no messages
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    aprsint._schedule_dedup_cleanup()

    # A clean-up should not have been scheduled.
    eq_(len(ax25int._loop.calls), 0)

def test_schedule_dedup_cleanup_pending():
    """
    Test _schedule_dedup_cleanup cancels existing pending clean-ups
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a pending clean-up
    deduplication_timeout = aprsint._loop.call_later(1, None)
    aprsint._deduplication_timeout = deduplication_timeout

    # Schedule the next one
    aprsint._schedule_dedup_cleanup()

    # We should now no-longer have a pending clean-up
    assert deduplication_timeout.cancelled()
    assert_is(aprsint._deduplication_timeout, None)

def test_schedule_dedup_cleanup_oldest_future():
    """
    Test _schedule_dedup_cleanup schedules for expiry of oldest message
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a few hashes
    now = aprsint._loop.time()
    aprsint._msg_expiry.update({
        b'hash1': now + 1,
        b'hash2': now + 2,
        b'hash3': now + 3
    })

    # Schedule the clean-up
    aprsint._schedule_dedup_cleanup()

    # A clean-up should have been scheduled.
    eq_(len(ax25int._loop.calls), 1)
    (calltime, callfunc) = ax25int._loop.calls.pop(0)

    # Should be scheduled for the earliest expiry
    assert_less(calltime - now, 1.01)
    assert_greater(calltime - now, 0.99)
    eq_(callfunc, aprsint._dedup_cleanup)

def test_schedule_dedup_cleanup_oldest_past():
    """
    Test _schedule_dedup_cleanup schedules immediately for expired messages.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a few hashes
    now = aprsint._loop.time()
    aprsint._msg_expiry.update({
        b'hash1': now - 1,
        b'hash2': now + 2,
        b'hash3': now + 3
    })

    # Schedule the clean-up
    aprsint._schedule_dedup_cleanup()

    # A clean-up should have been scheduled.
    eq_(len(ax25int._loop.calls), 1)
    (calltime, callfunc) = ax25int._loop.calls.pop(0)

    # Should be scheduled pretty much now
    assert_less(calltime - now, 0.01)
    eq_(callfunc, aprsint._dedup_cleanup)

def test_dedup_cleanup_expired():
    """
    Test _dedup_cleanup removes only expired messages
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a few hashes
    now = aprsint._loop.time()
    aprsint._msg_expiry.update({
        b'hash1': now - 1,
        b'hash2': now + 2,
        b'hash3': now + 3
    })

    # Perform the clean-up
    aprsint._dedup_cleanup()

    # We should no longer have 'hash1'
    assert b'hash1' not in aprsint._msg_expiry

    # but should have the others
    assert b'hash2' in aprsint._msg_expiry
    assert b'hash3' in aprsint._msg_expiry

    # There should be a re-schedule pending
    eq_(len(ax25int._loop.calls), 1)
    (calltime, callfunc) = ax25int._loop.calls.pop(0)

    # Should be scheduled pretty much now
    assert_less(calltime - now, 0.01)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

def test_on_receive_exception():
    """
    Test _on_receive swallows exceptions.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                pid=0xf0,
                payload=b':VK4BWI-2 :testing{123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Stub the _test_or_add_frame so it returns false
    aprsint._test_or_add_frame = lambda *a : False

    # Stub the IOLoop's call_soon so it fails
    calls = []
    def stub(*args):
        calls.append(args)
        raise RuntimeError('Oopsie')
    aprsint._loop.call_soon = stub

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # We should have called call_soon, but the exception should have been
    # caught and logged.
    eq_(len(calls), 1)

def test_on_receive_notframe():
    """
    Test _on_receive ignores non-APRS-frames.
    """
    # Create a frame
    class DummyFrame(AX25UnnumberedInformationFrame):
        def __init__(self, *args, **kwargs):
            self.addressee_calls = 0
            super(DummyFrame, self).__init__(*args, **kwargs)

        @property
        def addressee(self):
            self.addressee_calls += 1
            return AX25Address.decode('N0CALL')


    frame = DummyFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                pid=0xf0,
                payload=b'this is not an APRS message',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Stub the _test_or_add_frame so it returns false
    aprsint._test_or_add_frame = lambda *a : False

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # The addressee property should not be touched
    eq_(frame.addressee_calls, 0)

def test_on_receive_dup():
    """
    Test _on_receive ignores duplicate frames.
    """
    # Create a frame and hash it
    frame = AX25UnnumberedInformationFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                pid=0xf0,
                payload=b':VK4BWI-2 :testing{123',
                repeaters=['WIDE2-1','WIDE1-1']
    )
    framedigest = APRSInterface._hash_frame(frame)

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject the hash
    now = aprsint._loop.time()
    aprsint._msg_expiry.update({
        framedigest: now + 3
    })

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be no calls made
    eq_(len(ax25int._loop.calls), 0)

def test_on_receive_pass_to_router():
    """
    Test _on_receive passes the message to the base APRSRouter class.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                pid=0xf0,
                payload=b':VK4BWI-2 :testing{123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be two calls made, one to our deduplication clean-up, the
    # other to our superclass
    eq_(len(ax25int._loop.calls), 2)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, super(APRSInterface, aprsint)._on_receive)

def test_on_receive_addressed():
    """
    Test _on_receive of message addressed to station.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='APZAIO',
                source='VK4BWI-2',
                pid=0xf0,
                payload=b':VK4MSL-10:testing{123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be three calls made, one to our deduplication clean-up, the
    # other to our superclass, the third to our received_address_msg signal.
    eq_(len(ax25int._loop.calls), 3)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, super(APRSInterface, aprsint)._on_receive)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, aprsint.received_addressed_msg.emit)

def test_on_receive_unsol_ackrej():
    """
    Test _on_receive of unsolicited ACK/REJ addressed to station.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='APZAIO',
                source='VK4BWI-2',
                pid=0xf0,
                payload=b':VK4MSL-10:ack123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be two calls made, one to our deduplication clean-up, the
    # other to our superclass.  We don't pass the message out otherwise.
    eq_(len(ax25int._loop.calls), 2)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, super(APRSInterface, aprsint)._on_receive)

def test_on_receive_sol_ackrej():
    """
    Test _on_receive of solicited ACK/REJ addressed to station.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='APZAIO',
                source='VK4BWI-2',
                pid=0xf0,
                payload=b':VK4MSL-10:ack123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a message handler for the message ID
    handler = DummyMessageHandler()
    aprsint._pending_msg['123'] = handler

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be two calls made, one to our deduplication clean-up, the
    # other to our superclass.  The message will be passed to the handler.
    eq_(len(ax25int._loop.calls), 3)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, super(APRSInterface, aprsint)._on_receive)

    (_, callfunc, msg) = ax25int._loop.calls.pop(0)
    eq_(callfunc, handler._on_response)
    eq_(bytes(frame), bytes(msg))

def test_on_msg_handler_finish():
    """
    Test that _on_msg_handler_finish removes a message from the pending list
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a message handler for the message ID
    handler = DummyMessageHandler()
    aprsint._pending_msg['123'] = handler

    # Call the clean-up function
    aprsint._on_msg_handler_finish('123')

    # This should now be empty
    eq_(aprsint._pending_msg, {})



#!/usr/bin/env python3

import logging

from nose.tools import eq_

from aioax25.aprs.message import APRSMessageAckFrame
from aioax25.frame import AX25UnnumberedInformationFrame
from aioax25.aprs.router import APRSRouter

from ..loop import DummyLoop

def test_get_destination_msgframe():
    """
    Test _get_destination returns the addressee of a APRSMessageFrame.
    """
    frame = APRSMessageAckFrame(
            destination='APZAIO',
            addressee='VK4BWI-2',
            source='VK4MSL-7',
            msgid='123'
    )
    router = APRSRouter()
    destination = router._get_destination(frame)
    eq_(destination, frame.addressee)

def test_get_destination_uiframe():
    """
    Test _get_destination returns the destination field of a generic UI frame
    """
    frame = AX25UnnumberedInformationFrame(
            destination='APZAIO',
            source='VK4MSL-7',
            pid=0xf0,
            payload=b':BLN4     :Test bulletin'
    )
    router = APRSRouter()
    destination = router._get_destination(frame)
    eq_(destination, frame.header.destination)



#!/usr/bin/env python3

"""
AX.25 framing
"""

import re
from collections.abc import Sequence

# Frame type classes

class AX25Frame(object):
    """
    Base class for AX.25 frames.
    """

    POLL_FINAL  = 0b00010000

    CONTROL_I_MASK  = 0b00000001
    CONTROL_I_VAL   = 0b00000000
    CONTROL_US_MASK = 0b00000011
    CONTROL_S_VAL   = 0b00000001
    CONTROL_U_VAL   = 0b00000011

    # PID codes
    PID_ISO8208_CCITT   = 0x01
    PID_VJ_IP4_COMPRESS = 0x06
    PID_VJ_IP4          = 0x07
    PID_SEGMENTATION    = 0x08
    PID_TEXNET          = 0xc3
    PID_LINKQUALITY     = 0xc4
    PID_APPLETALK       = 0xca
    PID_APPLETALK_ARP   = 0xcb
    PID_ARPA_IP4        = 0xcc
    PID_APRA_ARP        = 0xcd
    PID_FLEXNET         = 0xce
    PID_NETROM          = 0xcf
    PID_NO_L3           = 0xf0
    PID_ESCAPE          = 0xff

    @classmethod
    def decode(cls, data):
        """
        Decode a single AX.25 frame from the given data.
        """
        (header, data) = AX25FrameHeader.decode(data)
        if not data:
            raise ValueError('Insufficient packet data')

        # Next should be the control field
        control = data[0]
        data = data[1:]

        if (control & cls.CONTROL_I_MASK) == cls.CONTROL_I_VAL:
            # This is an I frame - TODO
            #return AX25InformationFrame.decode(header, control, data)
            return AX25RawFrame(
                    destination=header.destination,
                    source=header.source,
                    repeaters=header.repeaters,
                    cr=header.cr,
                    control=control,
                    payload=data
            )
        elif (control & cls.CONTROL_US_MASK) == cls.CONTROL_S_VAL:
            # This is a S frame - TODO
            #return AX25SupervisoryFrame.decode(header, control, data)
            return AX25RawFrame(
                    destination=header.destination,
                    source=header.source,
                    repeaters=header.repeaters,
                    cr=header.cr,
                    control=control,
                    payload=data
            )
        elif (control & cls.CONTROL_US_MASK) == cls.CONTROL_U_VAL:
            # This is a U frame
            return AX25UnnumberedFrame.decode(header, control, data)
        else: # pragma: no cover
            # This should not happen because all possible bit combinations
            # are covered above.
            assert False, 'How did we get here?'

    def __init__(self, destination, source, repeaters=None, cr=False):
        self._header = AX25FrameHeader(destination, source, repeaters, cr)

    def _encode(self):
        """
        Generate the encoded AX.25 frame.
        """
        # Send the addressing header
        for byte in bytes(self.header):
            yield byte

        # Send the control byte
        yield self._control

        # Send the payload
        for byte in self.frame_payload:
            yield byte

    def __bytes__(self):
        """
        Encode the AX.25 frame
        """
        return bytes(self._encode())

    def __str__(self):
        return str(self._header)

    @property
    def header(self):
        return self._header

    @property
    def control(self):
        return self._control

    @property
    def frame_payload(self):
        """
        Return the bytes in the frame payload (following the control byte)
        """
        return b''

    def copy(self, header=None):
        """
        Make a copy of this frame with a new header for digipeating.
        """
        clone = self._copy()
        if header is not None:
            clone._header = header
        return clone


class AX25RawFrame(AX25Frame):
    """
    A representation of a raw AX.25 frame.
    """

    def __init__(self, destination, source, control, repeaters=None,
            cr=False, payload=None):
        self._header = AX25FrameHeader(destination, source, repeaters, cr)
        self._control = control
        self._payload = payload or b''

    @property
    def frame_payload(self):
        return self._payload

    def _copy(self):
        return self.__class__(
                destination=self.header.destination,
                source=self.header.source,
                control=self.control,
                repeaters=self.header.repeaters,
                cr=self.header.cr,
                payload=self.frame_payload
        )


class AX25UnnumberedFrame(AX25Frame):
    """
    A representation of an un-numbered frame.
    """
    MODIFIER_MASK = 0b11101111

    @classmethod
    def decode(cls, header, control, data):
        # Decode based on the control field
        modifier = control & cls.MODIFIER_MASK
        for subclass in (
                AX25UnnumberedInformationFrame,
                AX25FrameRejectFrame
        ):
            if modifier == subclass.MODIFIER:
                return subclass.decode(header, control, data)

        # If we're still here, clearly this is a plain U frame.
        if data:
            raise ValueError('Unnumbered frames (other than UI and '\
                            'FRMR) do not have payloads')

        return cls(
                destination=header.destination,
                source=header.source,
                repeaters=header.repeaters,
                cr=header.cr,
                modifier=modifier,
                pf=bool(control & cls.POLL_FINAL)
        )

    def __init__(self, destination, source, modifier,
            repeaters=None, pf=False, cr=False):
        super(AX25UnnumberedFrame, self).__init__(
                destination=destination, source=source,
                repeaters=repeaters, cr=cr)
        self._pf = bool(pf)
        self._modifier = int(modifier) & self.MODIFIER_MASK

    @property
    def _control(self):
        """
        Return the value of the control byte.
        """
        control = self._modifier
        if self._pf:
            control |= self.POLL_FINAL
        return control

    @property
    def pf(self):
        """
        Return the state of the poll/final bit
        """
        return self._pf

    @property
    def modifier(self):
        """
        Return the modifier bits
        """
        return self._modifier

    def _copy(self):
        return self.__class__(
                destination=self.header.destination,
                source=self.header.source,
                repeaters=self.header.repeaters,
                modifier=self.modifier,
                cr=self.header.cr,
                pf=self.pf
        )


class AX25UnnumberedInformationFrame(AX25UnnumberedFrame):
    """
    A representation of an un-numbered information frame.
    """
    MODIFIER = 0b00000011

    @classmethod
    def decode(cls, header, control, data):
        if not data:
            raise ValueError('Payload of UI must be at least one byte')
        return cls(
                destination=header.destination,
                source=header.source,
                repeaters=header.repeaters,
                cr=header.cr,
                pf=bool(control & cls.POLL_FINAL),
                pid=data[0],
                payload=data[1:]
        )

    def __init__(self, destination, source, pid, payload,
            repeaters=None, pf=False, cr=False):
        super(AX25UnnumberedInformationFrame, self).__init__(
                destination=destination, source=source,
                repeaters=repeaters, cr=cr, pf=pf,
                modifier=self.MODIFIER)
        self._pid = int(pid) & 0xff
        self._payload = bytes(payload)

    @property
    def pid(self):
        return self._pid

    @property
    def payload(self):
        return self._payload

    @property
    def frame_payload(self):
        return bytearray([self.pid]) + self.payload

    def __str__(self):
        return '%s: PID=0x%02x Payload=%r' % (
                self.header,
                self.pid,
                self.payload)

    def _copy(self):
        return self.__class__(
                destination=self.header.destination,
                source=self.header.source,
                repeaters=self.header.repeaters,
                cr=self.header.cr,
                pf=self.pf,
                pid=self.pid,
                payload=self.payload
        )


class AX25FrameRejectFrame(AX25UnnumberedFrame):
    """
    A representation of a Frame Reject (FRMR) frame.

    Not much effort has been made to decode the meaning of these bits.
    """

    MODIFIER = 0b10000111
    W_MASK   = 0b00000001
    X_MASK   = 0b00000010
    Y_MASK   = 0b00000100
    Z_MASK   = 0b00001000
    VR_MASK  = 0b11100000
    VR_POS   = 5
    CR_MASK  = 0b00010000
    VS_MASK  = 0b00001110
    VS_POS   = 1


    @classmethod
    def decode(cls, header, control, data):
        if len(data) != 3:
            raise ValueError('Payload of FRMR must be 3 bytes')

        # W, X, Y and Z bits
        w = bool(data[0] & cls.W_MASK)
        x = bool(data[0] & cls.X_MASK)
        y = bool(data[0] & cls.Y_MASK)
        z = bool(data[0] & cls.Z_MASK)

        # VR, CR and VS fields
        vr = (data[1] & cls.VR_MASK) >> cls.VR_POS
        cr = bool(data[1] & cls.CR_MASK)
        vs = (data[1] & cls.VS_MASK) >> cls.VS_POS

        # Control field of rejected frame
        frmr_control = data[2]

        return cls(
                destination=header.destination,
                source=header.source,
                repeaters=header.repeaters,
                cr=header.cr,
                pf=bool(control & cls.POLL_FINAL),
                w=w, x=x, y=y, z=z,
                vr=vr, frmr_cr=cr, vs=vs,
                frmr_control=frmr_control
        )

    def __init__(self, destination, source, w, x, y, z,
            vr, frmr_cr, vs, frmr_control,
            repeaters=None, pf=False, cr=False):
        super(AX25FrameRejectFrame, self).__init__(
                destination=destination, source=source,
                repeaters=repeaters, cr=cr, pf=pf,
                modifier=self.MODIFIER)

        self._w = bool(w)
        self._x = bool(x)
        self._y = bool(y)
        self._z = bool(z)
        self._frmr_cr = bool(frmr_cr)
        self._frmr_control = int(frmr_control)
        self._vr = int(vr)
        self._vs = int(vs)

    @property
    def frame_payload(self):
        return bytes(self._gen_frame_payload())

    def _gen_frame_payload(self):
        wxyz = 0
        if self._w:
            wxyz |= self.W_MASK
        if self._x:
            wxyz |= self.X_MASK
        if self._y:
            wxyz |= self.Y_MASK
        if self._z:
            wxyz |= self.Z_MASK
        yield wxyz

        vrcrvs = 0
        vrcrvs |= (self._vr << self.VR_POS) & self.VR_MASK
        if self._frmr_cr:
            vrcrvs |= self.CR_MASK
        vrcrvs |= (self._vs << self.VS_POS) & self.VS_MASK
        yield vrcrvs

        yield self._frmr_control

    @property
    def w(self):
        return self._w

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    @property
    def vs(self):
        return self._vs

    @property
    def vr(self):
        return self._vr

    @property
    def frmr_cr(self):
        return self._frmr_cr

    @property
    def frmr_control(self):
        return self._frmr_control

    def _copy(self):
        return self.__class__(
                destination=self.header.destination,
                source=self.header.source,
                repeaters=self.header.repeaters,
                w=self.w, x=self.x, y=self.y, z=self.z,
                frmr_cr=self.frmr_cr, vr=self.vr, vs=self.vs,
                frmr_control=self.frmr_control,
                cr=self.header.cr, pf=self.pf
        )


# Helper classes

class AX25FrameHeader(object):
    """
    A representation of an AX.25 frame header.
    """
    @classmethod
    def decode(cls, data):
        """
        Decode a frame header from the data given, return the
        decoded header and the data remaining.
        """
        # Decode the addresses
        addresses = []
        while data and (not (addresses and addresses[-1].extension)):
            addresses.append(AX25Address.decode(data))
            data = data[7:]

        # Whatever's left is the frame payload data.
        if len(addresses) < 2:
            raise ValueError('Too few addresses')

        return (cls(
            destination=addresses[0],
            source=addresses[1],
            repeaters=addresses[2:],
            cr=addresses[0].ch
        ), data)

    def __init__(self, destination, source, repeaters=None,
            cr=False):
        self._cr = bool(cr)
        self._destination = AX25Address.decode(destination)
        self._source = AX25Address.decode(source)
        self._repeaters = AX25Path(*(repeaters or []))

    def _encode(self):
        """
        Generate an encoded AX.25 frame header
        """
        # Extension bit should be 0
        # CH bit should be 1 for command, 0 for response
        self._destination.extension = False
        self._destination.ch = self._cr
        for byte in bytes(self._destination):
            yield byte

        # Extension bit should be 0 if digipeaters follow, 1 otherwise
        # CH bit should be 0 for command, 1 for response
        self._source.extension = not bool(self._repeaters)
        self._source.ch = not self._cr
        for byte in bytes(self._source):
            yield byte

        # Digipeaters
        if self._repeaters:
            # Very last should have extension = 1, others extension 0.
            # Leave the CH bits untouched.
            for rpt in self._repeaters[:-1]:
                rpt.extension = False
            self._repeaters[-1].extension = True

            for rpt in self._repeaters:
                for byte in bytes(rpt):
                    yield byte

    def __str__(self):
        """
        Dump the frame header in human-readable form.
        """
        return '%s>%s%s' % (
                self._source,
                self._destination,
                (',%s' % self._repeaters) \
                        if self._repeaters else ''
        )

    def __bytes__(self):
        """
        Encode the AX.25 frame header
        """
        return bytes(self._encode())

    @property
    def destination(self):
        return self._destination

    @property
    def source(self):
        return self._source

    @property
    def repeaters(self):
        return self._repeaters

    @property
    def cr(self):
        return self._cr


class AX25Path(Sequence):
    """
    A representation of a digipeater path.
    """
    def __init__(self, *path):
        """
        Construct a path using the given path.
        """
        self._path = tuple([
            AX25Address.decode(digi)
            for digi in path
        ])

    def __len__(self):
        """
        Return the path length
        """
        return len(self._path)

    def __getitem__(self, index):
        """
        Return the Nth path element.
        """
        return self._path[index]

    def __str__(self):
        """
        Return a string representation of the digipeater path.
        """
        return ','.join(
                str(addr) for addr in self._path
        )

    def __repr__(self):
        """
        Return the Python representation of the digipeater path.
        """
        return '%s(%s)' % (
                self.__class__.__name__,
                ', '.join([repr(addr) for addr in self._path])
        )

    @property
    def reply(self):
        """
        Return the reply path (the "consumed" digipeaters in reverse order).
        """
        return self.__class__(
                *tuple([
                    digi.copy(ch=False)
                    for digi in
                    filter(
                        lambda digi : digi.ch,
                        reversed(self._path)
                    )
                ])
        )

    def replace(self, alias, address):
        """
        Replace an address alias (e.g. WIDE1-1) with the given address
        (e.g. the address of this station).
        """
        alias = AX25Address.decode(alias).normalised
        address = AX25Address.decode(address)
        return self.__class__(
                *tuple([
                    address if (digi.normalised == alias) else digi
                    for digi in self._path
                ])
        )


class AX25Address(object):
    """
    A representation of an AX.25 address (callsign + SSID)
    """

    CALL_RE = re.compile(r'^([0-9A-Z]+)(?:-([0-9]{1,2}))?(\*?)$')

    @classmethod
    def decode(cls, data):
        """
        Decode an AX.25 address from a frame.
        """
        if isinstance(data, (bytes, bytearray)):
            # Ensure the data is at least 7 bytes!
            if len(data) < 7:
                raise ValueError('AX.25 addresses must be 7 bytes!')

            # This is a binary representation in the AX.25 frame header
            callsign = bytes([
                b >> 1
                for b in data[0:6]
            ]).decode('US-ASCII').strip()
            ssid        = (data[6]          & 0b00011110) >> 1
            ch          = bool(data[6]      & 0b10000000)
            res1        = bool(data[6]      & 0b01000000)
            res0        = bool(data[6]      & 0b00100000)
            extension   = bool(data[6]      & 0b00000001)
            return cls(callsign, ssid, ch, res0, res1, extension)
        elif isinstance(data, str):
            # This is a human-readable representation
            match = cls.CALL_RE.match(data.upper())
            if not match:
                raise ValueError('Not a valid SSID: %s' % data)
            return cls(
                    callsign=match.group(1),
                    ssid=int(match.group(2) or 0),
                    ch=match.group(3) == '*'
            )
        elif isinstance(data, AX25Address):
            # Clone factory
            return data.copy()
        else:
            raise TypeError("Don't know how to decode %r" % data)

    def __init__(self, callsign, ssid=0,
            ch=False, res0=True, res1=True, extension=False):
        self._callsign  = str(callsign).upper()
        self._ssid      = int(ssid) & 0b00001111
        self._ch        = bool(ch)
        self._res0      = bool(res0)
        self._res1      = bool(res1)
        self._extension = bool(extension)

    def _encode(self):
        """
        Generate the encoded AX.25 address.
        """
        for byte in self._callsign[0:6].ljust(6).encode('US-ASCII'):
            yield (byte << 1)

        # SSID byte
        ssid = self._ssid << 1
        if self._extension:
            ssid |= 0b00000001
        if self._res0:
            ssid |= 0b00100000
        if self._res1:
            ssid |= 0b01000000
        if self._ch:
            ssid |= 0b10000000
        yield ssid

    def __bytes__(self):
        """
        Return the encoded call-sign.
        """
        return bytes(self._encode())

    def __str__(self):
        """
        Return the call-sign and SSID as a string.
        """
        address = self.callsign
        if self.ssid > 0:
            address += '-%d' % self.ssid

        if self.ch:
            address += '*'
        return address

    def __repr__(self):
        """
        Return the Python representation of this object.
        """
        return ('%s(callsign=%s, ssid=%d, ch=%r, res0=%r, '\
                'res1=%r, extension=%r)') % (\
                self.__class__.__name__,
                self.callsign, self.ssid, self.ch,
                self.res0, self.res1, self.extension
        )

    def __eq__(self, other):
        if not isinstance(other, AX25Address):
            return NotImplemented

        for field in ('callsign', 'ssid', 'extension',
                    'res0', 'res1', 'ch'):
            if getattr(self, field) != getattr(other, field):
                return False

        return True

    def __hash__(self):
        return hash(tuple([
            getattr(self, field)
            for field in
            ('callsign', 'ssid', 'extension', 'res0', 'res1', 'ch')
        ]))

    @property
    def callsign(self):
        return self._callsign

    @property
    def ssid(self):
        return self._ssid

    @property
    def extension(self):
        return self._extension

    @extension.setter
    def extension(self, value):
        self._extension = bool(value)

    @property
    def res0(self):
        return self._res0

    @property
    def res1(self):
        return self._res1

    @property
    def ch(self):
        return self._ch

    @ch.setter
    def ch(self, value):
        self._ch = bool(value)

    def copy(self, **overrides):
        """
        Return a copy of this address, optionally with fields overridden.
        """
        mydata = dict(
                callsign=self.callsign,
                ssid=self.ssid,
                ch=self.ch,
                res0=self.res0,
                res1=self.res1,
                extension=self.extension
        )
        mydata.update(overrides)
        return self.__class__(**mydata)

    @property
    def normalised(self):
        """
        Return a normalised copy of this address.  (Set reserved bits to ones,
        clear the CH bit and extension bit.)
        """
        return self.copy(res0=True, res1=True, ch=False, extension=False)
